#!/usr/bin/env python3

# Copyright (C) 2023, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import sys
import PyInstaller.__main__

src = "blackboard_sync"
is_win = sys.platform.startswith('win')
sep = ';' if is_win else ':'

icon = f'{src}/assets/logo'
icon += '.ico' if is_win else '.png'

assets = f'{src}/assets'
assets = assets + sep + assets

qt = f'{src}/qt'
qt = qt + sep + qt

uni = f'{src}/universities.json'
uni += sep + src

PyInstaller.__main__.run([
    f'--icon={icon}',
    f'--add-data={assets}',
    f'--add-data={qt}',
    f'--add-data={uni}',
    '--noconfirm',
    '--clean',
    '--windowed',
    '--name=BBSync',
    'main.py',
])
